@@ -6,11 +6,9 @@
 
 
 import asyncio
-import errno
 import os
 import os.path
 import platform
-import random
 import re
 import shutil
 import socket
@@ -36,29 +34,15 @@ else:
         return name
 
 
-def find_available_port(port_range=(49152, 65535), max_tries=1000):
-    low, high = port_range
-
-    port = low
-    try_no = 0
-
-    while try_no < max_tries:
-        try_no += 1
-        port = random.randint(low, high)
-        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
-        try:
-            sock.bind(('127.0.0.1', port))
-        except socket.error as e:
-            if e.errno == errno.EADDRINUSE:
-                continue
-        finally:
-            sock.close()
-
-        break
-    else:
-        port = None
-
-    return port
+def find_available_port():
+    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
+    try:
+        sock.bind(('127.0.0.1', 0))
+        return sock.getsockname()[1]
+    except Exception:
+        return None
+    finally:
+        sock.close()
 
 
 class ClusterError(Exception):
@@ -69,7 +53,10 @@ class Cluster:
     def __init__(self, data_dir, *, pg_config_path=None):
         self._data_dir = data_dir
         self._pg_config_path = pg_config_path
-        self._pg_bin_dir = os.environ.get('PGINSTALLATION')
+        self._pg_bin_dir = (
+            os.environ.get('PGINSTALLATION')
+            or os.environ.get('PGBIN')
+        )
         self._pg_ctl = None
         self._daemon_pid = None
         self._daemon_process = None
@@ -124,6 +111,10 @@ class Cluster:
                 'cluster in {!r} has already been initialized'.format(
                     self._data_dir))
 
+        settings = dict(settings)
+        if 'encoding' not in settings:
+            settings['encoding'] = 'UTF-8'
+
         if settings:
             settings_args = ['--{}={}'.format(k, v)
                              for k, v in settings.items()]
@@ -164,8 +155,8 @@ class Cluster:
         sockdir = server_settings.get('unix_socket_directories')
         if sockdir is None:
             sockdir = server_settings.get('unix_socket_directory')
-        if sockdir is None:
-            sockdir = '/tmp'
+        if sockdir is None and _system != 'Windows':
+            sockdir = tempfile.gettempdir()
 
         ssl_key = server_settings.get('ssl_key_file')
         if ssl_key:
@@ -176,12 +167,13 @@ class Cluster:
             server_settings = server_settings.copy()
             server_settings['ssl_key_file'] = keyfile
 
-        if self._pg_version < (9, 3):
-            sockdir_opt = 'unix_socket_directory'
-        else:
-            sockdir_opt = 'unix_socket_directories'
+        if sockdir is not None:
+            if self._pg_version < (9, 3):
+                sockdir_opt = 'unix_socket_directory'
+            else:
+                sockdir_opt = 'unix_socket_directories'
 
-        server_settings[sockdir_opt] = sockdir
+            server_settings[sockdir_opt] = sockdir
 
         for k, v in server_settings.items():
             extra_args.extend(['-c', '{}={}'.format(k, v)])
@@ -193,6 +185,14 @@ class Cluster:
             # privileges.
             if os.getenv('ASYNCPG_DEBUG_SERVER'):
                 stdout = sys.stdout
+                print(
+                    'asyncpg.cluster: Running',
+                    ' '.join([
+                        self._pg_ctl, 'start', '-D', self._data_dir,
+                        '-o', ' '.join(extra_args)
+                    ]),
+                    file=sys.stderr,
+                )
             else:
                 stdout = subprocess.DEVNULL
 
@@ -521,7 +521,10 @@ class Cluster:
 
     def _find_pg_config(self, pg_config_path):
         if pg_config_path is None:
-            pg_install = os.environ.get('PGINSTALLATION')
+            pg_install = (
+                os.environ.get('PGINSTALLATION')
+                or os.environ.get('PGBIN')
+            )
             if pg_install:
                 pg_config_path = platform_exe(
                     os.path.join(pg_install, 'pg_config'))
@@ -642,7 +645,7 @@ class HotStandbyCluster(TempCluster):
         if self._pg_version >= (12, 0):
             server_settings = server_settings.copy()
             server_settings['primary_conninfo'] = (
-                'host={host} port={port} user={user}'.format(
+                '"host={host} port={port} user={user}"'.format(
                     host=self._master['host'],
                     port=self._master['port'],
                     user=self._repl_user,
