@@ -8,6 +8,11 @@
 
 #include "recordobj.h"
 
+#ifdef _PyObject_GC_IS_TRACKED
+# define _ApgObject_GC_IS_TRACKED _PyObject_GC_IS_TRACKED
+#else
+# define _ApgObject_GC_IS_TRACKED PyObject_GC_IsTracked
+#endif
 
 static PyObject * record_iter(PyObject *);
 static PyObject * record_new_items_iter(PyObject *);
@@ -15,31 +20,54 @@ static PyObject * record_new_items_iter(PyObject *);
 static ApgRecordObject *free_list[ApgRecord_MAXSAVESIZE];
 static int numfree[ApgRecord_MAXSAVESIZE];
 
+static size_t MAX_RECORD_SIZE = (
+    ((size_t)PY_SSIZE_T_MAX - sizeof(ApgRecordObject) - sizeof(PyObject *))
+    / sizeof(PyObject *)
+);
+
 
 PyObject *
-ApgRecord_New(PyObject *desc, Py_ssize_t size)
+ApgRecord_New(PyTypeObject *type, PyObject *desc, Py_ssize_t size)
 {
     ApgRecordObject *o;
     Py_ssize_t i;
+    int need_gc_track = 0;
 
     if (size < 0 || desc == NULL || !ApgRecordDesc_CheckExact(desc)) {
         PyErr_BadInternalCall();
         return NULL;
     }
 
-    if (size < ApgRecord_MAXSAVESIZE && (o = free_list[size]) != NULL) {
-        free_list[size] = (ApgRecordObject *) o->ob_item[0];
-        numfree[size]--;
-        _Py_NewReference((PyObject *)o);
-    }
-    else {
-        /* Check for overflow */
-        if ((size_t)size > ((size_t)PY_SSIZE_T_MAX - sizeof(ApgRecordObject) -
-                    sizeof(PyObject *)) / sizeof(PyObject *)) {
+    if (type == &ApgRecord_Type) {
+        if (size < ApgRecord_MAXSAVESIZE && (o = free_list[size]) != NULL) {
+            free_list[size] = (ApgRecordObject *) o->ob_item[0];
+            numfree[size]--;
+            _Py_NewReference((PyObject *)o);
+        }
+        else {
+            /* Check for overflow */
+            if ((size_t)size > MAX_RECORD_SIZE) {
+                return PyErr_NoMemory();
+            }
+            o = PyObject_GC_NewVar(ApgRecordObject, &ApgRecord_Type, size);
+            if (o == NULL) {
+                return NULL;
+            }
+        }
+
+        need_gc_track = 1;
+    } else {
+        assert(PyType_IsSubtype(type, &ApgRecord_Type));
+
+        if ((size_t)size > MAX_RECORD_SIZE) {
             return PyErr_NoMemory();
         }
-        o = PyObject_GC_NewVar(ApgRecordObject, &ApgRecord_Type, size);
-        if (o == NULL) {
+        o = (ApgRecordObject *)type->tp_alloc(type, size);
+        if (!_ApgObject_GC_IS_TRACKED((PyObject *)o)) {
+            PyErr_SetString(
+                PyExc_TypeError,
+                "record subclass is not tracked by GC"
+            );
             return NULL;
         }
     }
@@ -51,7 +79,9 @@ ApgRecord_New(PyObject *desc, Py_ssize_t size)
     Py_INCREF(desc);
     o->desc = (ApgRecordDescObject*)desc;
     o->self_hash = -1;
-    PyObject_GC_Track(o);
+    if (need_gc_track) {
+        PyObject_GC_Track(o);
+    }
     return (PyObject *) o;
 }
 
@@ -201,26 +231,39 @@ record_richcompare(PyObject *v, PyObject *w, int op)
     Py_ssize_t vlen, wlen;
     int v_is_tuple = 0;
     int w_is_tuple = 0;
+    int v_is_record = 0;
+    int w_is_record = 0;
     int comp;
 
-    if (!ApgRecord_CheckExact(v)) {
-        if (!PyTuple_Check(v)) {
-            Py_RETURN_NOTIMPLEMENTED;
-        }
+    if (PyTuple_Check(v)) {
         v_is_tuple = 1;
     }
-
-    if (!ApgRecord_CheckExact(w)) {
-        if (!PyTuple_Check(w)) {
-            Py_RETURN_NOTIMPLEMENTED;
-        }
-        w_is_tuple = 1;
+    else if (ApgRecord_CheckExact(v)) {
+        v_is_record = 1;
+    }
+    else if (!ApgRecord_Check(v)) {
+        Py_RETURN_NOTIMPLEMENTED;
     }
 
+    if (PyTuple_Check(w)) {
+        w_is_tuple = 1;
+    }
+    else if (ApgRecord_CheckExact(w)) {
+        w_is_record = 1;
+    }
+    else if (!ApgRecord_Check(w)) {
+        Py_RETURN_NOTIMPLEMENTED;
+    }
+
+
 #define V_ITEM(i) \
-    (v_is_tuple ? (PyTuple_GET_ITEM(v, i)) : (ApgRecord_GET_ITEM(v, i)))
+    (v_is_tuple ? \
+        PyTuple_GET_ITEM(v, i) \
+        : (v_is_record ? ApgRecord_GET_ITEM(v, i) : PySequence_GetItem(v, i)))
 #define W_ITEM(i) \
-    (w_is_tuple ? (PyTuple_GET_ITEM(w, i)) : (ApgRecord_GET_ITEM(w, i)))
+    (w_is_tuple ? \
+        PyTuple_GET_ITEM(w, i) \
+        : (w_is_record ? ApgRecord_GET_ITEM(w, i) : PySequence_GetItem(w, i)))
 
     vlen = Py_SIZE(v);
     wlen = Py_SIZE(w);
@@ -520,7 +563,7 @@ record_values(PyObject *o, PyObject *args)
 static PyObject *
 record_keys(PyObject *o, PyObject *args)
 {
-    if (!ApgRecord_CheckExact(o)) {
+    if (!ApgRecord_Check(o)) {
         PyErr_BadInternalCall();
         return NULL;
     }
@@ -532,7 +575,7 @@ record_keys(PyObject *o, PyObject *args)
 static PyObject *
 record_items(PyObject *o, PyObject *args)
 {
-    if (!ApgRecord_CheckExact(o)) {
+    if (!ApgRecord_Check(o)) {
         PyErr_BadInternalCall();
         return NULL;
     }
@@ -544,7 +587,7 @@ record_items(PyObject *o, PyObject *args)
 static int
 record_contains(ApgRecordObject *o, PyObject *arg)
 {
-    if (!ApgRecord_CheckExact(o)) {
+    if (!ApgRecord_Check(o)) {
         PyErr_BadInternalCall();
         return -1;
     }
@@ -660,7 +703,7 @@ record_iter_next(ApgRecordIterObject *it)
     seq = it->it_seq;
     if (seq == NULL)
         return NULL;
-    assert(ApgRecord_CheckExact(seq));
+    assert(ApgRecord_Check(seq));
 
     if (it->it_index < Py_SIZE(seq)) {
         item = ApgRecord_GET_ITEM(seq, it->it_index);
@@ -716,7 +759,7 @@ record_iter(PyObject *seq)
 {
     ApgRecordIterObject *it;
 
-    if (!ApgRecord_CheckExact(seq)) {
+    if (!ApgRecord_Check(seq)) {
         PyErr_BadInternalCall();
         return NULL;
     }
@@ -774,7 +817,7 @@ record_items_next(ApgRecordItemsObject *it)
     if (seq == NULL) {
         return NULL;
     }
-    assert(ApgRecord_CheckExact(seq));
+    assert(ApgRecord_Check(seq));
     assert(it->it_key_iter != NULL);
 
     key = PyIter_Next(it->it_key_iter);
@@ -854,7 +897,7 @@ record_new_items_iter(PyObject *seq)
     ApgRecordItemsObject *it;
     PyObject *key_iter;
 
-    if (!ApgRecord_CheckExact(seq)) {
+    if (!ApgRecord_Check(seq)) {
         PyErr_BadInternalCall();
         return NULL;
     }
