@@ -13,7 +13,6 @@ import tempfile
 
 import asyncpg
 from asyncpg import _testbase as tb
-from asyncpg import compat
 
 
 class TestCopyFrom(tb.ConnectedTestCase):
@@ -467,7 +466,6 @@ class TestCopyTo(tb.ConnectedTestCase):
                 def __init__(self):
                     self.rowcount = 0
 
-                @compat.aiter_compat
                 def __aiter__(self):
                     return self
 
@@ -507,7 +505,6 @@ class TestCopyTo(tb.ConnectedTestCase):
                 def __init__(self):
                     self.rowcount = 0
 
-                @compat.aiter_compat
                 def __aiter__(self):
                     return self
 
@@ -533,7 +530,6 @@ class TestCopyTo(tb.ConnectedTestCase):
                 def __init__(self):
                     self.rowcount = 0
 
-                @compat.aiter_compat
                 def __aiter__(self):
                     return self
 
@@ -564,7 +560,6 @@ class TestCopyTo(tb.ConnectedTestCase):
                     self.rowcount = 0
                     self.loop = loop
 
-                @compat.aiter_compat
                 def __aiter__(self):
                     return self
 
@@ -649,6 +644,29 @@ class TestCopyTo(tb.ConnectedTestCase):
         finally:
             await self.con.execute('DROP TABLE copytab')
 
+    async def test_copy_records_to_table_async(self):
+        await self.con.execute('''
+            CREATE TABLE copytab_async(a text, b int, c timestamptz);
+        ''')
+
+        try:
+            date = datetime.datetime.now(tz=datetime.timezone.utc)
+            delta = datetime.timedelta(days=1)
+
+            async def record_generator():
+                for i in range(100):
+                    yield ('a-{}'.format(i), i, date + delta)
+
+                yield ('a-100', None, None)
+
+            res = await self.con.copy_records_to_table(
+                'copytab_async', records=record_generator())
+
+            self.assertEqual(res, 'COPY 101')
+
+        finally:
+            await self.con.execute('DROP TABLE copytab_async')
+
     async def test_copy_records_to_table_no_binary_codec(self):
         await self.con.execute('''
             CREATE TABLE copytab(a uuid);
