@@ -24,11 +24,7 @@ from asyncpg import pool as pg_pool
 _system = platform.uname().system
 
 
-if os.environ.get('TRAVIS_OS_NAME') == 'osx':
-    # Travis' macOS is _slow_.
-    POOL_NOMINAL_TIMEOUT = 0.5
-else:
-    POOL_NOMINAL_TIMEOUT = 0.1
+POOL_NOMINAL_TIMEOUT = 0.1
 
 
 class SlowResetConnection(pg_connection.Connection):
@@ -243,7 +239,7 @@ class TestPool(tb.ConnectedTestCase):
             with self.assertRaisesRegex(
                     asyncpg.InterfaceError,
                     r'cannot call Cursor\.forward.*released '
-                    r'back to the pool'.format(meth=meth)):
+                    r'back to the pool'):
 
                 c.forward(1)
 
@@ -379,6 +375,26 @@ class TestPool(tb.ConnectedTestCase):
             self.cluster.trust_local_connections()
             self.cluster.reload()
 
+    async def test_pool_handles_task_cancel_in_acquire_with_timeout(self):
+        # See https://github.com/MagicStack/asyncpg/issues/547
+        pool = await self.create_pool(database='postgres',
+                                      min_size=1, max_size=1)
+
+        async def worker():
+            async with pool.acquire(timeout=100):
+                pass
+
+        # Schedule task
+        task = self.loop.create_task(worker())
+        # Yield to task, but cancel almost immediately
+        await asyncio.sleep(0.00000000001)
+        # Cancel the worker.
+        task.cancel()
+        # Wait to make sure the cleanup has completed.
+        await asyncio.sleep(0.4)
+        # Check that the connection has been returned to the pool.
+        self.assertEqual(pool._queue.qsize(), 1)
+
     async def test_pool_handles_task_cancel_in_release(self):
         # Use SlowResetConnectionPool to simulate
         # the Task.cancel() and __aexit__ race.
@@ -704,6 +720,27 @@ class TestPool(tb.ConnectedTestCase):
 
         await pool.close()
 
+    async def test_pool_size_and_capacity(self):
+        async with self.create_pool(
+            database='postgres',
+            min_size=2,
+            max_size=3,
+        ) as pool:
+            self.assertEqual(pool.get_min_size(), 2)
+            self.assertEqual(pool.get_max_size(), 3)
+            self.assertEqual(pool.get_size(), 2)
+            self.assertEqual(pool.get_idle_size(), 2)
+
+            async with pool.acquire():
+                self.assertEqual(pool.get_idle_size(), 1)
+
+                async with pool.acquire():
+                    self.assertEqual(pool.get_idle_size(), 0)
+
+                    async with pool.acquire():
+                        self.assertEqual(pool.get_size(), 3)
+                        self.assertEqual(pool.get_idle_size(), 0)
+
     @unittest.skipIf(sys.version_info[:2] < (3, 6), 'no asyncgen support')
     async def test_pool_handles_transaction_exit_in_asyncgen_1(self):
         pool = await self.create_pool(database='postgres',
@@ -826,6 +863,7 @@ class TestPool(tb.ConnectedTestCase):
             await pool.release(con)
 
         self.assertIsNone(pool._holders[0]._con)
+        await pool.close()
 
     async def test_pool_set_connection_args(self):
         pool = await self.create_pool(database='postgres',
@@ -867,6 +905,8 @@ class TestPool(tb.ConnectedTestCase):
         con = await pool.acquire()
         self.assertEqual(con.get_settings().application_name,
                          'set_conn_args_test_2')
+        await pool.release(con)
+        await pool.close()
 
     async def test_pool_init_race(self):
         pool = self.create_pool(database='postgres', min_size=1, max_size=1)
